CON
' 
' Specify the type of this plugin (must be unique):
'
TOGGLE = 1
'
' Specify the size of any Hub RAM block to be allocated (this block contains
' one long that specifies the pins to be flashed - this long can be updated
' in real-time - the plugin checks it before each flash to determine the pins
' to turn on.
'
BLOCK_SIZE = 4
'
' Specify the services that this plugin will accept:
'
INITIALIZE   = 1
GET_BUFFER   = 2
START_TOGGLE = 3
STOP_TOGGLE  = 4
SET_RATE     = 5
'
' Define the services accepted by this plugin (examples only):
'
'name: INITIALIZE  
'code: 1
'type: short request
'data: buffer address (to update with current state)
'rslt: 0 

'name: GET_BUFFER
'code: 2
'type: short request
'data: none
'rslt: -1 if not initialized, otherwise buffer address

'name: START_TOGGLE
'code: 3
'type: short request
'data: none
'rslt: -1 if not initialized, otherwise 0 

'name: STOP_TOGGLE
'code: 4
'type: short request
'data: none
'rslt: -1 if not initialized, otherwise 0       

'name: SET_RATE
'code: 5
'type: long request
'data: delay (in clock ticks) between toggling outputs
'rslt: -1 if not initialized, otherwise 0
'
OBJ
'
' Include common definitions
'
   Common : "Common.Spin"
'
VAR
'
' This variable is only used during initialization of the plugin
'
   long PLUGIN_BLOCK

CON
''==============================================================================
' LAYER 0 METHODS
''==============================================================================
PUB Setup
'
' Setup the plugun - first allocate the necessary Hub RAM 
'
  PLUGIN_BLOCK := long[Common#FREE_MEM] - BLOCK_SIZE
  long[Common#FREE_MEM] := PLUGIN_BLOCK
'
' set no pins toggled by default
'
  long[PLUGIN_BLOCK] := 0

PUB Start | cog
'
' This methood loads and starts the plugin.  
' Returns -1 on error, or the cog used
'   
  cog := cognew(@entry, Common#REGISTRY)
   
  if (cog => 0)
    ' wait till plugin has registered itself
    repeat until long[Common#REGISTRY][cog] >> 24 == TOGGLE
    ' send initialization data to the plugin    
    Common.ShortPluginRequest(TOGGLE, INITIALIZE, PLUGIN_BLOCK)
  cog += 1
'
DAT
              org     0
'
' entry point for the cog program
'               
entry

'---------------------------------- REGISTRATION ------------------------------
'
' this code is executed only once - after execution it can be used as variable space
'
rep           COGID   rep         ' re becomes a pointer to our registry entry
rqp           SHL     rep,#2      ' multiply our cogid by 4
rsp           ADD     rep,PAR     ' add the registry address (passed in PAR)
rqst          RDLONG  rqp,rep     ' read our registry entry 
rslt          SHL     rqp,#8      ' update ...
buffp         OR      rqp,#TOGGLE ' ... registry ...
rate          ROR     rqp,#8      ' ... usage ...
state         WRLONG  rqp,rep     ' ... bits
pins          MOV     rsp,rqp     ' set up response pointer ...
time          ADD     rsp,#4      ' ... as second long of comms block
t1            MOV     buffp,#0    ' indicate we are not initialized yet
t2            RDLONG  rate,#0     ' set toggle rate to clock speed (i.e. 1Hz)
t3            NEG     state,#1    ' start with state = -1 (toggling disabled)

'-------------------------------- REQUEST DECODING ---------------------------
'
' decode requests and process them, or do other stuff if there are no requests
'
get_request
              RDLONG    rqst,rqp wz             ' any requests?
        if_z  JMP       #do_stuff               ' no - do other stuff!
              MOV       t1,rqst                 ' yes - get ...
              SHR       t1,#24                  ' ... request id
              CMPS      t1,#SET_RATE wz,wc      ' valid request id?
        if_a  JMP       #return_err             ' no - return error
              ADD       t1,#(rqst_table-1)      ' fetch ...
              MOVS      :table,t1               ' ... service ...
              NOP                               ' ... (nop required here!) ...
:table        MOV       t2,0-0                  ' ... address
              JMP       t2                      ' jump to service
'
' table of request entry points
'
rqst_table
              long      _Initialize             ' Request 1 entry point         
              long      _Get_Buffer             ' Request 2 entry point         
              long      _Start_Toggle           ' Request 3 entry point 
              long      _Stop_Toggle            ' Request 4 entry point 
              long      _Set_Rate               ' Request 5 entry point 
'             ... etc ...
'
' common request exit points
'        
return_err
              NEG       rslt,#1                 ' set result to -1 if error
              JMP       #end_request
return_ok
              MOV       rslt,#0                 ' set result to 0 if ok
end_request
              WRLONG    rslt,rsp                ' write result to comms block 
done_request
              WRLONG    zero,rqp                ' clear request block
              JMP       #get_request            ' process next request 

'
' we can do stuff while waiting for a request - for this plugin this is where we
' do the actual toggling of the outputs (if enabled). Note that we always turn
' off the same pins we turned on, and only update the pins to be toggled each 
' time we turn them on.
'
do_stuff
              TJZ       buffp,#get_request      ' don't do anything if not initialized   
              MOV       t1,cnt                  ' get current time
              MOV       t2,t1                   ' remember it
              MOV       t3,time                 ' get previous time
              SUB       t2,t3                   ' calculate time since last toggle
              CMPS      t2,rate wz,wc           ' time to toggle again?
        if_b  JMP       #get_request            ' no - continue processing requests
              MOV       time,t1                 ' yes - update previous time
              CMPS      state,#0 wz,wc          ' check toggle state
        if_b  JMP       #get_request            ' toggling is disabled

        if_nz ANDN      outa,pins               ' last state was on, so set pins off
        if_nz MOV       state,#0                ' remember new state
        if_nz JMP       #get_request            ' done
              
              RDLONG    t1,buffp                ' get new pins to toggle
              CMP       pins,t1 wz              ' same as old pins?
        if_nz ANDN      dira,pins               ' no - update ...
        if_nz OR        dira,t1                 ' ... output pins ...
        if_nz MOV       pins,t1                 ' ... and remember new pins
              OR        outa,pins               ' set pins on
              MOV       state,#1                ' remember new state
              JMP       #get_request


'---------------------------- REQUEST HANDLING ROUTINES ------------------------
'
' _Initialize - accept a pointer to our allocated buffer.
'
_Initialize
              MOV       buffp,rqst              ' save the buffer pointer
              JMP       #return_ok              ' done
'
' _GetBuffer - return the pointer to our allocated buffer (-1 if not initialized)
'
_Get_Buffer
              MOV       rslt,buffp              ' get the buffer pointer
              JMP       #end_request            ' done
'
' _Set_Rate - set the rate (number of clock ticks) between toggling outputs
'
_Set_Rate
              RDLONG    rate,rqst               ' lower 24 bits is address of rate
              JMP       #return_ok              ' done
'
' _Start_Toggle - start togglong outputs at the specified rate   
'
_Start_Toggle
              MOV       time,cnt                ' save the system counter 
              OR        outa,pins               ' set pins low
              OR        dira,pins               ' set pins to outputs
              MOV       state,#0                ' state = 0 or 1 means enable toggling
              JMP       #end_request
'
' _Stop_Toggle - stop toggling outputs   
'
_Stop_Toggle
              MOV       time,cnt                ' save the system counter 
              ANDN      outa,pins               ' yes - set pins low
              ANDN      dira,pins               ' set pins to inputs
              NEG       state,#1                ' state = -1 means disable toggling
              JMP       #end_request
'
zero          long      0                       ' 
'
              fit       $1f0
